# Copyright (c) 2012 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.


import logging
import os
import shlex
import sys
import tempfile
import time

from pylib import android_commands
from pylib import constants
from pylib import pexpect
from pylib.android_commands import errors

from test_package import TestPackage


class TestPackageApk(TestPackage):
  """A helper class for running APK-based native tests.

  Args:
    adb: ADB interface the tests are using.
    device: Device to run the tests.
    test_suite: A specific test suite to run, empty to run all.
    timeout: Timeout for each test.
    cleanup_test_files: Whether or not to cleanup test files on device.
    tool: Name of the Valgrind tool.
  """

  def __init__(self, adb, device, test_suite, timeout,
               cleanup_test_files, tool, apk_package_name,
               test_activity_name, command_line_file):
    TestPackage.__init__(self, adb, device, test_suite, timeout,
                         cleanup_test_files, tool)
    self._apk_package_name = apk_package_name
    self._test_activity_name = test_activity_name
    self._command_line_file = command_line_file

  def _CreateTestRunnerScript(self, options):
    command_line_file = tempfile.NamedTemporaryFile()
    # GTest expects argv[0] to be the executable path.
    command_line_file.write(self.test_suite_basename + ' ' + options)
    command_line_file.flush()
    self.adb.PushIfNeeded(command_line_file.name,
                          constants.TEST_EXECUTABLE_DIR + '/' +
                          self._command_line_file)

  def _GetGTestReturnCode(self):
    return None

  def _GetFifo(self):
    # The test.fifo path is determined by:
    # testing/android/java/src/org/chromium/native_test/
    #     ChromeNativeTestActivity.java and
    # testing/android/native_test_launcher.cc
    return '/data/data/' + self._apk_package_name + '/files/test.fifo'

  def _ClearFifo(self):
    self.adb.RunShellCommand('rm -f ' + self._GetFifo())

  def _WatchFifo(self, timeout, logfile=None):
    for i in range(10):
      if self.adb.FileExistsOnDevice(self._GetFifo()):
        logging.info('Fifo created.')
        break
      time.sleep(i)
    else:
      raise errors.DeviceUnresponsiveError(
          'Unable to find fifo on device %s ' % self._GetFifo())
    args = shlex.split(self.adb.Adb()._target_arg)
    args += ['shell', 'cat', self._GetFifo()]
    return pexpect.spawn('adb', args, timeout=timeout, logfile=logfile)

  def ClearApplicationState(self):
    """Clear the application state."""
    self.adb.ClearApplicationState(self._apk_package_name)

  def GetAllTests(self):
    """Returns a list of all tests available in the test suite."""
    self._CreateTestRunnerScript('--gtest_list_tests')
    try:
      self.tool.SetupEnvironment()
      # Clear and start monitoring logcat.
      self._ClearFifo()
      self.adb.RunShellCommand(
          'am start -n ' + self._apk_package_name + '/' +
          self._test_activity_name)
      # Wait for native test to complete.
      p = self._WatchFifo(timeout=30 * self.tool.GetTimeoutScale())
      p.expect("<<ScopedMainEntryLogger")
      p.close()
    finally:
      self.tool.CleanUpEnvironment()
    # We need to strip the trailing newline.
    content = [line.rstrip() for line in p.before.splitlines()]
    ret = self._ParseGTestListTests(content)
    return ret

  def CreateTestRunnerScript(self, gtest_filter, test_arguments):
    self._CreateTestRunnerScript('--gtest_filter=%s %s' % (gtest_filter,
                                                           test_arguments))

  def RunTestsAndListResults(self):
    try:
      self.tool.SetupEnvironment()
      self._ClearFifo()
      self.adb.RunShellCommand(
        'am start -n ' + self._apk_package_name + '/' +
        self._test_activity_name)
    finally:
      self.tool.CleanUpEnvironment()
    logfile = android_commands.NewLineNormalizer(sys.stdout)
    return self._WatchTestOutput(self._WatchFifo(timeout=10, logfile=logfile))

  def _NeedsInstall(self):
    pm_path_output = self.adb.RunShellCommand(
        'pm path ' + self._apk_package_name)
    if not pm_path_output:
      return True
    # pm_path_output is of the form: "package:/path/to/foo.apk"
    installed_apk_path = pm_path_output[0].split(':')[1]
    return not self.adb.CheckMd5Sum(
        self.test_suite_full, installed_apk_path, ignore_paths=True)

  def StripAndCopyExecutable(self):
    self.tool.CopyFiles()
    if self._NeedsInstall():
      # Always uninstall the previous one (by activity name); we don't
      # know what was embedded in it.
      self.adb.ManagedInstall(self.test_suite_full, False,
                              package_name=self._apk_package_name)

  def _GetTestSuiteBaseName(self):
    """Returns the  base name of the test suite."""
    # APK test suite names end with '-debug.apk'
    return os.path.basename(self.test_suite).rsplit('-debug', 1)[0]
